require "colorize"
require "string_scanner"

module Hence
  struct Microcode
    def initialize(@runner : Emulator::Data ->)
    end

    def initialize(&@runner : Emulator::Data ->)
    end

    def run(data : Emulator::Data) : Emulator::Data
      @runner.call(data)

      data
    end

    MICROCODES = {
      0x00_u8 => Microcode.new { },   # nop
      0x01_u8 => Microcode.new do |d| # dbg
        stack = if d.reg_sp == 0
                  [] of UInt16
                else
                  d.reg_sp.times.map_with_index { |_, i| d.get_memory({{ 32_u16 * 1024 }} + i) }
                end

        puts "#{"[DEBUG]".colorize(:red).mode(:bold)}: (#{d.reg_sp}) #{stack}"
        print "Press enter to continue execution...".colorize.mode(:dim)
        gets
      end,
      0x02_u8 => Microcode.new do |d| # jmp
        d.reg_mccpc = d.reg_tmp
      end,
      0x03_u8 => Microcode.new do |d| # jif
        if d.reg_tmp == 1
          d.reg_mccpc = d.reg_a
        end
      end,
      0x04_u8 => Microcode.new do |d| # tmpl
        d.set_register((d.reg_mccarg & 0xff).to_u8, d.reg_tmp)
      end,
      0x05_u8 => Microcode.new do |d| # tmplc
        d.set_register((d.reg_mccarg & 0xff).to_u8, d.reg_tmp) if d.reg_a == 1
      end,
      0x06_u8 => Microcode.new do |d| # tmps
        d.reg_tmp = d.reg_mccarg
      end,
      0x07_u8 => Microcode.new do |d| # tmpsr
        d.reg_tmp = d.get_register((d.reg_mccarg & 0xff).to_u8)
      end,
      0x08_u8 => Microcode.new do |d| # push
        d.set_memory({{ 32_u16 * 1024 }} + d.reg_sp, d.reg_tmp)
        d.reg_sp += 1
      end,
      0x09_u8 => Microcode.new do |d| # drop
        d.reg_tmp = d.get_memory({{ 32_u16 * 1024 }} + d.reg_sp - 1)
        d.set_memory({{ 32_u16 * 1024 }} + d.reg_sp - 1, 0_u16)
        d.reg_sp -= 1 if d.reg_sp.positive?
      end,
      0x0a_u8 => Microcode.new do |d| # pick
        d.reg_tmp = d.get_memory({{ 32_u16 * 1024 }} + d.reg_sp - 1 - d.reg_tmp)
      end,
      0x0b_u8 => Microcode.new do |d| # alu
        d.alu((d.reg_mccarg & 0xff).to_u8)
      end,
      0x0c_u8 => Microcode.new do |d| # load
        d.reg_tmp = d.get_memory(d.reg_tmp)
      end,
      0x0d_u8 => Microcode.new do |d| # store
        d.set_memory(d.reg_tmp, d.reg_a)
      end,
      0x0e_u8 => Microcode.new do |d| # print
        print d.reg_tmp
      end,
      0x0f_u8 => Microcode.new do |d| # emit
        print String.new(Bytes[d.reg_tmp])
      end,
    } of UInt8 => Microcode
    MICROCODE_NAMES = {
      "nop"   => 0x00_u8,
      "dbg"   => 0x01_u8,
      "jmp"   => 0x02_u8,
      "jif"   => 0x03_u8,
      "tmpl"  => 0x04_u8,
      "tmplc" => 0x05_u8,
      "tmps"  => 0x06_u8,
      "tmpsr" => 0x07_u8,
      "push"  => 0x08_u8,
      "drop"  => 0x09_u8,
      "pick"  => 0x0a_u8,
      "alu"   => 0x0b_u8,
      "load"  => 0x0c_u8,
      "store" => 0x0d_u8,
      "print" => 0x0e_u8,
      "emit"  => 0x0f_u8,
    } of String => UInt8
  end
end
